import os
import logging
from typing import List
import re
import pandas as pd
import subprocess
from io import StringIO
from abc import ABC, abstractmethod
import yaml

class SlurmScriptWriter(ABC):
    """
    This abstract base class is a template for script-writing classes that writes the body of Slurm scripts.
    """

    @abstractmethod
    def generate_script_body(self) -> List[str]:
        """
        Generates the body of the script. Returns a List of lines as strings.
        """
        pass

class MLPerfScriptWriter(SlurmScriptWriter):
    """
    Writes the script body for MLPerf scripts.
    """

    def __init__(self, 
                 run_id: int, 
                 benchmark: str, 
                 model: str, 
                 backend: str, 
                 arch: str,
                 container_image: str, 
                 data_path: str,
                 destination: str = None) -> None:
        """
        Initialize relevants variables for writing an MLPerf script.

        Args:
            run_id (int): Unique ID of this benchmark run
            benchmark (str): Type of benchmark
            model (str): Name of model (e.g. resnet50, bert99, etc.)
            backend (str): Backend of the model (e.g. tf, torch)
            arch (str): GPU architecture (e.g. gh200, h100, etc.)
            container_image (str): Path to pre-configured Apptainer image
            data_path (str): Path to dataset
            destination (str, optional): Path to where the script and files are located
        """
        # Get relevant environment variables
        self.PWD = os.environ.get('PWD')
        self.USER = os.environ.get('USER')

        self.RUN_ID = run_id

        # Validate + assign model parameters
        if not benchmark:
            raise ValueError("Invalid input for `benchmark`. Cannot be an empty string.")
        self.BENCHMARK = benchmark

        if not model:
            raise ValueError("Invalid input for `model`. Cannot be an empty string.")
        self.MODEL = model

        if not backend:
            raise ValueError("Invalid input for `backend`. Cannot be an empty string.")
        self.BACKEND = backend

        if not arch:
            raise ValueError("Invalid input for `arch`. Cannot be an empty string.")
        self.ARCH = arch

        # Initialize the directory for output files + directories
        # FIXME: Get rid of `destination` variable, if it becomes unnecessary
        if destination:
            self.DEST = destination
        else:
            self.DEST = self.PWD
        self.RESULTS_DIR = os.path.join(self.DEST, "results")
        self.SLURM_OUTPUT_DIR = os.path.join(self.DEST, "slurm_out")
        self.SLURM_ERROR_DIR = os.path.join(self.DEST, "slurm_err")
        self.LOG_DIR = os.path.join(self.DEST, "submit_log")
        self.SCRIPT_DIR = os.path.join(self.DEST, "scripts")
        self.OUTFILE = os.path.join(self.DEST, "test_results/default-reference-gpu-tf-v2.15.0-default_config/resnet50/offline/performance/run_1/mlperf_log_detail.txt")
        
        # Validate + assign path inputs
        # if not os.path.exists(container_image):
        #     raise ValueError("A valid container image must be specified.")
        self.CONTAINER_IMAGE = container_image

        # if not os.path.exists(data_path):
        #     raise ValueError("A valid path to dataset must be specified.")
        self.DATA_PATH = data_path

        # Initialize other relevant paths
        self.SCRAPE_METRICS_PATH = os.path.join(self.PWD, "scrape_metrics.py")
        self.YAML_PATH = os.path.join(self.PWD, "config.yaml")

        # Read the YAML file
        try:
            with open(self.YAML_PATH, 'r') as file:
                yaml_params = yaml.safe_load(file) # Returns None, if file is empty
        except OSError as e:
            raise e
        
        # Initialize model-specific variables
        model_config = yaml_params["model"][model]
        self.HW_NAME            = model_config["hw_name"]
        self.IMPLEMENTATION     = model_config["implementation"]
        self.DEVICE             = model_config["device"]
        self.SCENARIO           = model_config["scenario"]
        self.ADR_COMPILER_TAGS  = model_config["adr.compiler.tags"]
        self.TARGET_QPS         = model_config["target_qps"]
        self.CATEGORY           = model_config["category"]
        self.DIVISION           = model_config["division"]

    def get_slurm_args(self, args) -> None:
        """
        Read the slurm arguments and check if they are valid.
        """
        pass

    def make_dirs(self) -> None:
        """
        This function will check if folder exists or not,
        and create them. $PWD/results, $PWD/slurm_err,
        $PWD/slurm_out, $PWD/submit_log
        """
        # TODO: Log if folder exists.
        os.makedirs(self.RESULTS_DIR, exist_ok=True) # emulates 'mkdir -p' Linux command
        os.makedirs(self.SLURM_OUTPUT_DIR, exist_ok=True)
        os.makedirs(self.SLURM_ERROR_DIR, exist_ok=True)
        os.makedirs(self.LOG_DIR, exist_ok=True)
        os.makedirs(self.SCRIPT_DIR, exist_ok=True)
    
    def config_logger(self) -> logging.Logger:
        """
        This function configures the logger.
        """
        LOG_PATH = os.path.join(self.LOG_DIR, "submit_benchmarks_log.txt")
        logging.basicConfig(filename=LOG_PATH,
                            level=logging.INFO,
                            format="%(asctime)s - %(levelname)s: %(message)s",
                            datefmt="%Y-%m-%d %H:%M:%S")
        logger = logging.getLogger(__name__)
        return logger            
    
    def generate_script_body(self) -> List[str]:
        """
        This function will create the sbatch file.

        No arguments (e.g., number of nodes, model name, etc)
        should be specified inside this function. All of them
        should be inherited.
		"""
        return [
            # 'echo "Hello World"\n',
            'unset LD_LIBRARY_PATH\n',
            f'export APPTAINER_BINDPATH="/oscar/home/{self.USER}, /oscar/scratch/{self.USER}, /oscar/data"\n',
            'export APPTAINER_CACHEDIR=/tmp\n',
            '\n',
            'echo $SLURM_JOBID\n',
            '\n',
            f"srun apptainer exec --nv {self.CONTAINER_IMAGE} sh << 'EOF'\n" ,
            'export CM_REPOS=/tmp/CM\n',
            'cp -r /CM /tmp/.\n',
            '\n',
            f'cm run script "get validation dataset imagenet _2012 _full" --input={self.DATA_PATH}\n',
            f'cmr "run mlperf inference generate-run-cmds _submission" --quiet --submitter="MLCommons" --hw_name={self.HW_NAME} --model={self.MODEL} --implementation={self.IMPLEMENTATION} --backend={self.BACKEND} --device={self.DEVICE} --scenario={self.SCENARIO} --adr.compiler.tags={self.ADR_COMPILER_TAGS} --target_qps={self.TARGET_QPS} --category={self.CATEGORY} --division={self.DIVISION} --results_dir={self.RESULTS_DIR}\n',
            'EOF\n',
            '\n',
            # Move Slurm output files into directories
            f"mv *.out {self.SLURM_OUTPUT_DIR}\n",
            f"mv *.err {self.SLURM_ERROR_DIR}\n",
        ] # + self.scrape_metrics()
    
    def scrape_metrics(self) -> List[str]:
        """
        This function runs the scrape_metrics script.
        """
        return [
            # Save results to sqlite3 database
            f'python {self.SCRAPE_METRICS_PATH} {self.OUTFILE} --runid={self.RUN_ID} --benchmark={self.BENCHMARK} --model={self.MODEL} --backend={self.BACKEND} --arch={self.ARCH}\n'
        ]
        
    def slurm_input(self) -> List[int]:
        """
        Read input from slurm commands like sfeature and
        get the actual list of gpus to submit to.
        """
        # Create text file of the sfeature output
        try:
            result = subprocess.run(args=["sinfo", "-o", r"%N %P %c %m %f %G"], 
                                    shell=False, 
                                    capture_output=True, 
                                    text=True, 
                                    check=True) # check=True throws subprocess.CalledProcessError, if exits with code > 0
        except subprocess.CalledProcessError as e:
            print(e)

        # Read the output from sinfo into a pandas DataFrame
        df = pd.read_table(StringIO(result.stdout), delimiter=' ')

        # Extract the gpus
        gpus_df = df.loc[df["PARTITION"] == "gpu"]
        gpus_df.reset_index(drop=True, inplace=True)

        # Get the nodelist column
        nodelists = gpus_df["NODELIST"]
        node_list = []

        # Get a list of all the nodes
        pattern = re.compile(r'\[([^\[\]]+)\]') # Regex for strings enclosed in square brackets
        for nodelist in nodelists:
            stripped_node = nodelist.strip("gpu")
            match = pattern.search(stripped_node)
            if match:
                enclosed_string = match.group(1)
                items = enclosed_string.split(',')

                for item in items:
                    if '-' in item:
                        start, end = map(int, item.split('-'))
                        node_list.extend(range(start, end + 1))
                    else:
                        node_list.append(int(item))
            else:
                node_list.append(int(stripped_node))
        
        return node_list


"""
TEST: 
- 
"""

import pytest
from SlurmJobSubmitter import MLPerfJobSubmitter
import os

@pytest.fixture
def pwd():
    return os.environ.get("PWD")

@pytest.fixture
def script_body():
    return ['echo "Hello World"\n']

@pytest.fixture(autouse=True)
def test_script(pwd, script_body):
    """
    Submits a test script and returns the script path.
    """
    script_path = os.path.join(pwd, "test")
    example = MLPerfJobSubmitter(script_path=script_path,
                                 nodes=1,
                                 gres="gpu:1",
                                 ntasks_per_node=1,
                                 memory="40G",
                                 time="01:00:00",
                                 partition="gracehopper",
                                 account="ccv-gh200-gcondo",
                                 num_runs=1)
    print(len(script_body), script_body)
    example.submit(script_body)
    yield script_path
    os.remove(script_path)

# TEST: '#!/bin/bash is written at the top of the file
def test_submit_bash_annotation(test_script):
    with open(test_script, 'r') as file:
        file_lines = file.readlines()
        assert (file_lines[0] == "#!/bin/bash\n")

# TEST: All necessary SBATCH header lines are written
def test_submit_sbatch_header(test_script):
    correct_header = [
        '#SBATCH --nodes=1\n',
        '#SBATCH -p gracehopper\n',
        '#SBATCH --gres=gpu:1\n',
        '#SBATCH --account=ccv-gh200-gcondo\n',
        '#SBATCH --ntasks-per-node=1\n',
        '#SBATCH --mem=40G\n',
        '#SBATCH -t 01:00:00\n',
        '#SBATCH -e %j.err\n',
        '#SBATCH -o %j.out\n',
    ]
    with open(test_script, 'r') as file:
        file_lines = file.readlines()
        assert any(file_lines[i:i+len(correct_header)] == correct_header for i in range(len(file_lines) - len(correct_header)))

# TEST: Script body is successfully written to script
def test_submit_body(test_script, script_body):
    with open(test_script, 'r') as file:
        file_lines = file.readlines()
        assert any(file_lines[i:i+len(script_body)] == script_body for i in range(len(file_lines) - len(script_body) + 1))

# TEST: Output directories are successfully created
def test_output_folders_exist(pwd):
    assert os.path.exists(os.path.join(pwd, "results"))
    assert os.path.exists(os.path.join(pwd, "scripts"))
    assert os.path.exists(os.path.join(pwd, "slurm_err"))
    assert os.path.exists(os.path.join(pwd, "slurm_out"))
    assert os.path.exists(os.path.join(pwd, "submit_log"))


"""
This script creates a batch script that submits jobs to Slurm to run the MLPerf inference benchmarks.
"""

# python python-scripts/submit_resnet-50.py -d /oscar/data/ccvinter/mstu/gracehopper_eval/MLPerf_gpu/gh200_resnet-50 -i /oscar/data/shared/eval_gracehopper/container_images/MLPerf/arm64/mlperf-resnet-50-tf-arm64 --data /oscar/data/ccvinter/mstu/gracehopper_eval/data/imagenet/ILSVRC2012/val

# python python-scripts/submit_resnet-50.py -d /oscar/data/ccvinter/mstu/gracehopper_eval/MLPerf_gpu/gh200_resnet-50 -i /oscar/data/shared/eval_gracehopper/container_images/MLPerf/x86_64/mlperf-resnet-50-tf-x86_64 --data /oscar/data/ccvinter/mstu/gracehopper_eval/data/imagenet/ILSVRC2012/val

import os
import argparse
from oscar_benchmarking import SlurmJobSubmitter
from oscar_benchmarking import SlurmScriptGenerator
import pandas as pd
import yaml
from datetime import datetime

# Read the YAML file
try:
    with open("config.yaml", 'r') as file:
        general_params = yaml.safe_load(file) # Returns None, if file is empty
except OSError as e:
    raise e

# Parse the CSV
run_config_df = pd.read_csv("run_config.csv")

# Get the working directory + user from ENV
PWD = os.environ.get('PWD')
USER = os.environ.get('USER')

# Configure CLI
parser = argparse.ArgumentParser(prog="submit_resnet-50", description=__doc__)
parser.add_argument("-d", "--destination", 
                    dest="destination",
                    metavar="DIR",
                    help="Directory in which to run the script (default: '.')",
                    default=PWD)
args = parser.parse_args()

# Loop through each run configuration in the CSV
for _, run in run_config_df.iterrows():
    # Parameters for the current run
    RUN_ID = run["RUN_ID"]
    BENCHMARK = run["BENCHMARK"]
    MODEL = run["MODEL"]
    BACKEND = run["BACKEND"]
    ARCH = run["ARCH"]

    print(RUN_ID, BENCHMARK, MODEL, BACKEND, ARCH)

    # Get model args
    DESTINATION =       general_params["destination"]
    NUM_RUNS =          general_params["num_runs"]
    DATA_PATH =         general_params["model"][MODEL]["data_path"]
    CONTAINER_IMAGE =   general_params["arch"][ARCH]["container_image"]

    # Define the output script path
    BATCH_PATH = os.path.join(os.path.join(DESTINATION, "scripts"), f'{RUN_ID}_{BENCHMARK}_{MODEL}_{datetime.now().strftime("%Y%m%d")}.submit') # FIXME: refactor variable to remove os path joining

    slurm_script_writer = SlurmScriptGenerator.MLPerfScriptWriter(RUN_ID, BENCHMARK, MODEL, BACKEND, ARCH, container_image=CONTAINER_IMAGE, data_path=DATA_PATH)

    # Make directories necessary for this script to run
    slurm_script_writer.make_dirs()

    # Batch Variables
    NODES =             general_params["arch"][ARCH]["nodes"]
    PARTITION =         general_params["arch"][ARCH]["partition"]
    GRES =              general_params["arch"][ARCH]["gres"]
    ACCOUNT =           general_params["arch"][ARCH]["account"]
    NTASKS_PER_NODE =   general_params["arch"][ARCH]["ntasks_per_node"]
    MEMORY =            general_params["arch"][ARCH]["memory"]
    TIME =              general_params["arch"][ARCH]["time"]
    ERROR_FILE_NAME =   general_params["arch"][ARCH]["error_file_name"]
    OUTPUT_FILE_NAME =  general_params["arch"][ARCH]["output_file_name"]

    # Configure the logger
    logger = slurm_script_writer.config_logger()

    # Log important info
    logger.info("Command-line arguments:")
    for arg, val in vars(args).items():
        logger.info(f"{arg}: {val}")

    # Generate script body
    script_body = slurm_script_writer.generate_script_body()

    slurm_job_submitter = SlurmJobSubmitter.MLPerfJobSubmitter(script_path=BATCH_PATH,
                                        nodes=NODES, 
                                        gres=GRES, 
                                        ntasks_per_node=NTASKS_PER_NODE, 
                                        memory=MEMORY, 
                                        time=TIME, 
                                        partition=PARTITION, 
                                        error_file_name=ERROR_FILE_NAME, 
                                        output_file_name=OUTPUT_FILE_NAME, 
                                        account=ACCOUNT,
                                        num_runs=NUM_RUNS)
    slurm_job_submitter.submit(script_body)

    # Log successful submission
    logger.info("Batch script submitted successfully.\n")


# Scrape performance metrics from MLPerf inference results
# Metrics: 
# Accuracy, 
# Latency, 
# Throughput (results_samples_per_second), 
# Inference Time, 
# Memory Usage

import os
import re
import json
import sqlite3
import argparse
import logging

logging.basicConfig(filename='scrape_metrics_log.txt', 
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        )
logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(
        prog='PerfMetrics',
        description='Gets the MLPerf metrics from input folder and adds to sqlite3 database',
        epilog='Need to manually edit the metric for now')

parser.add_argument('file', help='path to mlperf_log_detail.txt')
parser.add_argument('--runid', help='ID of your benchmark run')
parser.add_argument('--benchmark', help='Name of the benchmark')
parser.add_argument('--model', help='Model name')
parser.add_argument('--backend', help='Backend used, pytorch, tf, etc')
parser.add_argument('--arch', help='Architecture, x86_64 or arm64')

# Get the command-line arguments
args = parser.parse_args()

# Set databae and outfile
database = os.path.abspath("/oscar/data/shared/eval_gracehopper/results/performance.sqlite3")
outfile = os.path.join(args.file)

# Log args
logger.info('Database path: %s\n' %str(database))
logger.info('Command-line arguments:')
for arg, val in vars(args).items():
    logger.info(f'{arg}: {val}')


# Write to database: update values here
def write_db(metric, value, database, args=args):
    # Database info
    perfmetrics = {}
    perfmetrics['RUN_ID'] = args.runid
    perfmetrics['BENCH_MARK'] = args.benchmark
    perfmetrics['ARCH'] = args.arch
    perfmetrics['MODEL'] = args.model
    perfmetrics['BACKEND'] = args.backend
    perfmetrics['METRIC'] = metric
    perfmetrics['VALUE'] = value
    
    # Connect to database
    conn = sqlite3.connect(database)
    cursor = conn.cursor()

    # Insert queries
    keys = ', '.join(perfmetrics.keys())
    placeholders = ', '.join(['?' for _ in range(len(perfmetrics))])
    values = tuple(perfmetrics.values())

    insert_query = f"INSERT INTO perfmetrics ({keys}) VALUES ({placeholders})"
    cursor.execute(insert_query, values)

    conn.commit()
    conn.close()



def parse_file(fname):
    data = []
    dict2 = {}
    with open(fname, 'r') as file:
        text = file.read()
        matches = re.findall(r':::MLLOG\s+(\{.*?\})', text)
        #return matches
        for match in matches:
            match += '}'
            entry = json.loads(match) # convert string to dictionary

            data.append(entry)

            key = entry.get('key')
            val = entry.get('value')
            if key and val:
                dict2[key] = val
    
    return dict2

# Get the results
results = parse_file(outfile)

# Write results to database
write_db('SAMPLES_PER_SECOND', results['result_samples_per_second'], database)
write_db('MIN_LATENCY_NS', results['result_min_latency_ns'], database)
write_db('MAX_LATENCY_NS', results['result_max_latency_ns'], database)
write_db('MEAN_LATENCY_NS', results['result_mean_latency_ns'], database)



"""
This script plots the performance metrics from the SQL database using matplotlib.

Requirements:
- pandas 2.2.2
- matplotlib 3.9.0
"""

import pandas as pd
import sqlite3
import os
import matplotlib.pyplot as plt
import argparse

DATABASE_PATH = os.path.abspath("/oscar/data/shared/eval_gracehopper/results/performance.sqlite3")

# Configure parser
parser = argparse.ArgumentParser(
    prog='PlotMetrics',
    description='Plots the MLPerf metrics from sqlite3 database'
)
parser.add_argument('--model', help='Model for which to plot metrics')

# Get the command-line arguments
args = parser.parse_args()

# Establish a connection to the sqlite3 db + read the db into a pandas dataframe
with sqlite3.connect(DATABASE_PATH) as dbcon:
    df = pd.read_sql_query("SELECT * FROM perfmetrics", dbcon)

# Convert metrics into columns for each run ID
run_metrics = df[['RUN_ID', 'METRIC', 'VALUE']].pivot(index='RUN_ID', columns='METRIC', values='VALUE')

# Group by run ID
unique_runs_gb = df[['BENCH_MARK', 'ARCH', 'MODEL', 'BACKEND']].groupby(df['RUN_ID'])
unique_runs = unique_runs_gb.aggregate({'BENCH_MARK': 'first', 'ARCH': 'first', 'MODEL': 'first', 'BACKEND': 'first'})

# Achieved desired table form to plot on X-Y plane
plot_df = unique_runs.join(run_metrics, on='RUN_ID')

def subplot_model_metrics(df: pd.DataFrame, x: str, args):
    """
    Plots the metrics for the model specified by the '--model' argument.
    If this argument doesn't exist, then input df should already contain one model.
    """
    figure, axis = plt.subplots(2,2)
    if args.model:
        df = df[df['MODEL'].str.contains(args.model)]
    df.plot(x=x, y='MAX_LATENCY_NS', ax=axis[0,0], kind='bar')
    axis[0,0].set_title("Max Latency (S)")
    axis[0,0].tick_params(axis='x', labelrotation=0)

    df.plot(x=x, y='MEAN_LATENCY_NS', ax=axis[0,1], kind='bar')
    axis[0,1].set_title("Mean Latency (S)")
    axis[0,1].tick_params(axis='x', labelrotation=0)

    df.plot(x=x, y='MIN_LATENCY_NS', ax=axis[1,0], kind='bar')
    axis[1,0].set_title("Min Latency (S)")
    axis[1,0].tick_params(axis='x', labelrotation=0)

    df.plot(x=x, y='SAMPLES_PER_SECOND', ax=axis[1,1], kind='bar')
    axis[1,1].set_title("Samples Per Second")
    axis[1,1].tick_params(axis='x', labelrotation=0)

# Plot the metrics for each run
if args.model:
    subplot_model_metrics(plot_df, 'ARCH', args)
else:
    # Otherwise, show the average of the model across all architectures for each metric
    model_gb_plot_df = plot_df[['MODEL', 'MAX_LATENCY_NS', 'MEAN_LATENCY_NS', 'MIN_LATENCY_NS', 'SAMPLES_PER_SECOND']].groupby(plot_df['MODEL']).agg({'MODEL': 'first', 'MAX_LATENCY_NS': 'mean', 'MEAN_LATENCY_NS': 'mean', 'MIN_LATENCY_NS': 'mean', 'SAMPLES_PER_SECOND': 'mean'})
    num_models = model_gb_plot_df['MODEL'].size
    subplot_model_metrics(model_gb_plot_df, 'MODEL', args)

# Adjust subplots for better viewing experience
plt.subplots_adjust(hspace=0.5)
# plt.subplot_tool() # Use this tool to adjust subplot spacing to your liking
plt.show()


import subprocess
from typing import List
from abc import ABC, abstractmethod

class SlurmJobSubmitter(ABC):
    """
    This class is an abstract base class for submitting jobs to the Slurm scheduler. 
    Example use cases: MLPerf jobs, generic gpu jobs, generic cpu jobs, array jobs, etc.

    NOTE: This class only submits jobs from ONE script. If you need to write another script, declare another submitter object.
    """
    def __init__(self,
                    script_path: str, 
                    nodes: int, 
                    gres: str, 
                    ntasks_per_node: int, 
                    memory: str, 
                    time: str, 
                    partition: str, 
                    error_file_name: str,
                    output_file_name: str) -> None:
        """
        Initialize variables necessary to write a SBATCH header.

        Args:
            script_path (str): File path to write to.
            nodes (int): The number of nodes to use.
            gres (str): Generic resources required for the job.
            ntasks_per_node (int): Number of tasks to run per node.
            memory (str): Amount of memory required.
            time (str): Maximum time the job can run.
            partition (str): The partition to submit the job to.
            error_file_name (str): Name of the file to which standard error will be written.
            output_file_name (str): Name of the file to which standard output will be written.
        """
        self.script_path = script_path
        self.nodes = nodes
        self.gres = gres
        self.ntasks_per_node = ntasks_per_node
        self.memory = memory
        self.time = time
        self.partition = partition
        self.error_file_name = error_file_name
        self.output_file_name = output_file_name
    
    @abstractmethod
    def write(self, sbatch_header: List[str], script_body: List[str]) -> None:
        """
        Writes the given SBATCH header and script body to a file.
        """
        pass

    @abstractmethod
    def submit(self, script_body: List[str]) -> None:
        """
        Submits the script to Slurm.
        """
        pass

class MLPerfJobSubmitter(SlurmJobSubmitter):
    """
    This class submits MLPerf jobs to the Slurm scheduler.

    NOTE: This class only submits jobs from ONE script. If you need to write another script, declare another submitter object.
    """
    def __init__(self,
                    script_path: str, 
                    nodes: int, 
                    gres: str, 
                    ntasks_per_node: int, 
                    memory: str, 
                    time: str, 
                    partition: str, 
                    error_file_name: str = "%j.err",
                    output_file_name: str = "%j.out", 
                    account: str = None,
                    num_runs: int = 1) -> None:
        """
        Initialize variables necessary to write a SBATCH header for a MLPerf script.

        Raises ValueError, if num_runs is less than 1.

        Args:
            script_path (str): File path to write to.
            nodes (int): The number of nodes to use.
            gres (str): Generic resources required for the job.
            ntasks_per_node (int): Number of tasks to run per node.
            memory (str): Amount of memory required.
            time (str): Maximum time the job can run.
            partition (str): The partition to submit the job to.
            error_file_name (str, optional): Name of the file to which standard error will be written. Defaults to "%j.err".
            output_file_name (str, optional): Name of the file to which standard output will be written. Defaults to "%j.out".
            account (str, optional): Account to be charged for the job. Defaults to None.
            num_runs (int, optional): Number of times to run the MLPerf script. Defaults to 1.
        """
        # Initialize the super class variables
        super().__init__(script_path, nodes, gres, ntasks_per_node, memory, time, partition, error_file_name, output_file_name)
        
        # Initialize the subclass variables
        self.account = account

        # Check the number of runs
        if num_runs < 1:
            raise ValueError("Number of runs must be at least 1.")
        self.num_runs = num_runs

        self.params = [
            nodes,
            partition,
            gres,
            account,
            ntasks_per_node,
            memory,
            time,
            error_file_name,
            output_file_name
        ]

        self.gpu_lines = [
            f'#SBATCH --nodes={nodes}\n',
            f'#SBATCH -p {partition}\n',
            f'#SBATCH --gres={gres}\n',
            f'#SBATCH --account={account}\n',
            f'#SBATCH --ntasks-per-node={ntasks_per_node}\n',
            f'#SBATCH --mem={memory}\n',
            f'#SBATCH -t {time}\n',
            f'#SBATCH -e {error_file_name}\n',
            f'#SBATCH -o {output_file_name}\n',
        ]
    
    def write(self, sbatch_header: List[str], script_body: List[str]) -> None:
        """
        Writes to a batch script, given a file path and script content.

        Raises OSError, if opening the file fails.
        """
        try:
            with open(self.script_path, 'w') as f:
                f.writelines(
                    sbatch_header 
                    + ['\n'] 
                    + script_body)
        except OSError as e:
            raise e
    
    def submit(self, script_body: List[str]) -> None:
        """
        Submits the script to gpu node(s)
        """
        sbatch_header = ['#!/bin/bash\n', '\n',]

        # If the parameter is valid, then append the respective line to the sbatch header
        for param, line in zip(self.params, self.gpu_lines):
            if param:
                sbatch_header.append(line)
        
        # If array job, add the array job line.
        if self.num_runs > 1:
            sbatch_header.append(f"#SBATCH --array 1-{self.num_runs}")
        
        # Write the batch script to file
        self.write(sbatch_header, script_body)

        print("Submitted job 123456.")
        # try:
        #     sbatch_result = subprocess.run(["sbatch", self.script_path], capture_output=True, text=True, check=True)
        # except subprocess.CalledProcessError as e:
        #     raise e
        # print(sbatch_result.stdout)
        